#include <windows.h>
#include <dbgeng.h>

void DllMain()
{
}

extern "C" HRESULT __stdcall DebugExtensionInitialize(PULONG Version, PULONG Flags)
{
    HRESULT hFuncRet=E_FAIL;

    PDEBUG_CONTROL4 control=0;
    PDEBUG_CLIENT4 client=0;
    if(DebugCreate(__uuidof(IDebugClient4), (PVOID *)&client) != S_OK)
        goto cleanup;
    if(client->QueryInterface(__uuidof(IDebugControl3), (PVOID *)&control) != S_OK)
        goto cleanup;

    control->Output(DEBUG_OUTPUT_NORMAL,"extension: initializing...\n");

    *Version = DEBUG_EXTENSION_VERSION(1, 0);
    *Flags = 0;

    hFuncRet = S_OK;

    cleanup:

    if(control)
        control->Release();
    if(client)
        client->Release();

    return hFuncRet;
}

extern "C" VOID __stdcall DebugExtensionUninitialize(void)
{
    PDEBUG_CONTROL4 control=0;
    PDEBUG_CLIENT4 client=0;
    if(DebugCreate(__uuidof(IDebugClient4), (PVOID *)&client) != S_OK)
        goto cleanup;
    if(client->QueryInterface(__uuidof(IDebugControl3), (PVOID *)&control) != S_OK)
        goto cleanup;

    control->Output(DEBUG_OUTPUT_NORMAL,"extension: uninitializing...\n");

    cleanup:

    if(control)
        control->Release();
    if(client)
        client->Release();
}

extern "C" VOID WINAPI DebugExtensionNotify(ULONG Notify, ULONG64 Argument)
{
    PDEBUG_CONTROL4 control=0;
    PDEBUG_CLIENT4 client=0;
    if(DebugCreate(__uuidof(IDebugClient4), (PVOID *)&client) != S_OK)
        goto cleanup;
    if(client->QueryInterface(__uuidof(IDebugControl3), (PVOID *)&control) != S_OK)
        goto cleanup;

    switch(Notify)
    {
        case DEBUG_NOTIFY_SESSION_ACTIVE:
            //control->Output(DEBUG_OUTPUT_NORMAL,"extension: got DEBUG_NOTIFY_SESSION_ACTIVE\n");
            break;

        case DEBUG_NOTIFY_SESSION_INACTIVE:
            //control->Output(DEBUG_OUTPUT_NORMAL,"extension: got DEBUG_NOTIFY_SESSION_INACTIVE\n");
            break;

        case DEBUG_NOTIFY_SESSION_ACCESSIBLE:
            //control->Output(DEBUG_OUTPUT_NORMAL,"extension: got DEBUG_NOTIFY_SESSION_ACCESSIBLE\n");
            break;

        case DEBUG_NOTIFY_SESSION_INACCESSIBLE:
            //control->Output(DEBUG_OUTPUT_NORMAL,"extension: got DEBUG_NOTIFY_SESSION_INACCESSIBLE\n");
            break;
    }

    cleanup:

    if(control)
        control->Release();
    if(client)
        client->Release();
}

HRESULT __stdcall bn(PDEBUG_CLIENT4 Client, PCSTR arg)
{
    typedef struct _MIRACLBIG {
        ULONG   len;
        PULONG  pDwords;
    } MIRACLBIG, *PMIRACLBIG;

    MIRACLBIG m;

    // VARS: general
    HRESULT hRes = E_FAIL;
    DWORD dwRet = 0;

    // VARS: other    
    UINT64 pBig = 0;
    UINT n_bytes = 0;
    PUCHAR bytes = 0;
    UINT n_ulongs = 0;
    PULONG ulongs = 0;
    BOOL b_negative = 0;

    // get interfaces
    //
    PDEBUG_CONTROL4 control = 0;
    PDEBUG_CLIENT4 client = 0;
    PDEBUG_DATA_SPACES3 dataspaces = 0;

    if(DebugCreate(__uuidof(IDebugClient4), (PVOID *)&client) != S_OK)
        goto cleanup;
    if(client->QueryInterface(__uuidof(IDebugControl3), (PVOID *)&control) != S_OK)
        goto cleanup;
    if(client->QueryInterface(__uuidof(IDebugDataSpaces3), (PVOID *)&dataspaces) != S_OK)
        goto cleanup;

    // get arg
    //
    DEBUG_VALUE pdv;

    memset(&pdv,0,sizeof(pdv));

    if(S_OK != control->Evaluate(arg, DEBUG_VALUE_INT64, &pdv, 0))
    {
        control->Output(DEBUG_OUTPUT_ERROR, "ERROR: evaluating argument\n");
        goto cleanup;
    }

    pBig = pdv.I64;

    control->Output(DEBUG_OUTPUT_NORMAL, "bignum @%08X\n", pBig);

    // read in the number structure
    // 
    if(S_OK != dataspaces->ReadVirtual(pBig, &m, (ULONG)sizeof(MIRACLBIG), &dwRet))
    {
        control->Output(DEBUG_OUTPUT_ERROR, "ERROR: reading memory from target @%08X\n", pBig);
        goto cleanup;
    }

    // pick off negative flag
    //
    b_negative = (m.len & 0x80000000) ? 1:0;

    m.len &= 0x7FFFFFFF;

    // verbose stuff
    //
    control->Output(DEBUG_OUTPUT_NORMAL, "    len=%d\n", m.len);
    control->Output(DEBUG_OUTPUT_NORMAL, "pdwords=%08X\n", m.pDwords);

    if(m.len > 64)
    {
        control->Output(DEBUG_OUTPUT_ERROR, "ERROR: over 128 coefficients, likely this is not a bignum!\n");
        goto cleanup;
    }

    // read in digits array
    //
    PBYTE pBytes = new UCHAR[m.len * 4];

    if(S_OK != dataspaces->ReadVirtual((ULONG64)m.pDwords, pBytes, m.len * 4, &dwRet))
    {
        control->Output(DEBUG_OUTPUT_ERROR, "ERROR: reading memory from target @\n", m.pDwords);
        goto cleanup;
    }

    // finally print them out
    //
    if(b_negative)
        control->Output(DEBUG_OUTPUT_NORMAL, "-");

    for(int i=m.len*4-1; i>=0; --i)
         control->Output(DEBUG_OUTPUT_NORMAL, "%02X", pBytes[i]);  
 
    control->Output(DEBUG_OUTPUT_NORMAL, "\n");

    // done!
    //
    hRes = S_OK;

    cleanup:

    if(pBytes)
        delete [] pBytes;
    if(dataspaces)
        dataspaces->Release();
    if(control)
        control->Release();
    if(client)
        client->Release();

    return hRes;
}

